package org.processmining.humbertogarciacaballero.models.model.layout.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.NodeParams;

import prefuse.data.Edge;
import prefuse.data.Node;

public class LayerContainer {

	public static final String LAYER_VALUE_HOLDER_KEY = "LAYER_VALUE_HOLDER_KEY";
	private List<Layer> layers;
	
	public LayerContainer() {
		layers = new ArrayList<>();
	}
	
	public LayerContainer(LayerContainer layerContainer) {
		this.layers = new ArrayList<>(layerContainer.layers.size());
		for (Layer layer : layerContainer.layers) {
			this.layers.add(new Layer(layer));
		}
	}
	
	public List<Node> getNodes(int layer) {
		return layers.get(layer).getNodes();
	}
	
	public int getSize(int layer) {
		if (layer >= layers.size()) return 0;
		return layers.get(layer).getSize();
	}
	
	public void addNode(Node n, int layer) {
		while (layers.size() <= layer) {
			layers.add(layers.size(), new Layer(layers.size()));
		}
		
		layers.get(layer).addNode(n);
		NodeParams p = (NodeParams)n.get(HierarchicalLayouter.NODE_PARAMS);
		p.layer = layer;
		p.rank = layers.get(layer).getSize() - 1;
	}
	
	public void moveNode(Node n, int newLayer) {
		removeNode(n);
		addNode(n, newLayer);
	}
	
	public void removeNode(Node n) {
		int oldLayer = getLayer(n);
		if (oldLayer != -1) {
			layers.get(oldLayer).remove(n);
		}
	}
	
	public int getLayer(Node n) {
		NodeParams p = (NodeParams)n.get(HierarchicalLayouter.NODE_PARAMS);
		return p.layer;
	}

	public void reset() {
		for (Layer layer : layers) {
			layer.clear();
		}
		layers.clear();
	}
	
	public int getSize() {
		return layers.size();
	}
	
/*	public List<Node> getPredecessors(Node node) {
		int layer = getLayer(node);
		List<Node> nodesLayer = new LinkedList<>(getNodes(layer - 1));
		List<Node> neighbors = new LinkedList<>(node.getNeighbors());
		
		nodesLayer.retainAll(neighbors);
		return nodesLayer;
	}*/
	
	public List<Layer> getLayers() {
		return layers;
	}
	
	public String toString() {
		return "[" +
				"layers:" + Arrays.toString(layers.toArray()) +
				"]";
				
	}

	public List<Node> getNeighbors(Node node, int nextLayer) {
		List<Node> result = new ArrayList<>();
		int layerToLook = getLayer(node) + nextLayer;
		if (layerToLook >= 0 && layerToLook < getSize()) {
			@SuppressWarnings("unchecked")
			Iterator<Node> neighbors = node.neighbors();
			while (neighbors.hasNext()) {
				Node n = neighbors.next();
				if (getLayer(n) == layerToLook){
					result.add(n);
				}
			}
			
			final List<Node> layerNodes = getNodes(layerToLook);
			Collections.sort(result, new Comparator<Node>() {
				@Override
				public int compare(Node o1, Node o2) {
					return Integer.compare(layerNodes.indexOf(o1), layerNodes.indexOf(o2));
				}
			});
		}
		
		return result;
	}
	
//	public List<Node> getNeighborsPlusSameLayer(Node node, int nextLayer) {
//		return 
//	}
	
	public int getDistance(Edge e) {
		return Math.abs(getLayer(e.getSourceNode()) - getLayer(e.getTargetNode()));
	}
	
	public int getMinimumLayer(Edge e) {
		return Math.min(getLayer(e.getSourceNode()), getLayer(e.getTargetNode()));
	}
	
	public int getMaximumLayer(Edge e) {
		return Math.max(getLayer(e.getSourceNode()), getLayer(e.getTargetNode()));
	}

	public int getIndex(Node node) {
		return getNodes(getLayer(node)).indexOf(node);
	}

	public int getHighestLayer() {
		return getSize() - 1;
	}

	public void print() {
		for (Layer layer : layers) {
			System.out.println("Layer " + layer.getPosition());
			for (Node node : layer.getNodes()) {
				System.out.print(node.getRow() + " [rank=" + ((NodeParams)node.get(HierarchicalLayouter.NODE_PARAMS)).rank + "] , ");
			}
			System.out.println();
		}
	}

    public void swap(int layer, int u, int v) {
        List<Node> nodes = getNodes(layer);
        Node nodeU = nodes.get(u);
        Node nodeV = nodes.get(v);
        ((NodeParams)nodeU.get(HierarchicalLayouter.NODE_PARAMS)).rank = v;
        ((NodeParams)nodeV.get(HierarchicalLayouter.NODE_PARAMS)).rank = u;
        
        Collections.swap(nodes, u, v);
    }

    public void reset(LayerContainer other) {
        reset();
        for (int i = 0; i < other.getSize(); i++) {
            List<Node> layerNodes = other.getNodes(i);
            for (Node node : layerNodes) {
                addNode(node, i);
            }
        }
    }
}
